#include <cstdio>

#define FUNCTIONS(FUNCTION) \
  FUNCTION(f, int v, int &r, int &&rr, const int &cr) \
  FUNCTION(g) \

#include "reverse_extension.h"

template<class Base>
struct Base0 : public Base
{
  Base0()
  {
    puts("Base0::Base0");
  }
  virtual ~Base0()
  {
    puts("Base0::~Base0");
  }
  int F(int v, int &r, int &&rr, const int &cr)
  {
    printf("Enter Base0::F(%d, %d, %d, %d)\n", v, r, rr, cr);
    if (int err = Base::f(v, r, std::move(rr), cr))
      return err;
    printf("Leave Base0::F(%d, %d, %d, %d)\n", v, r, rr, cr);
    return 0;
  }
};

template <class Base>
class Ext0 : public Base
{
public:
  Ext0()
  {
    puts("Ext0::Ext0");
  }
  virtual ~Ext0()
  {
    puts("Ext0::~Ext0");
  }
  //~ int f(int x) is not implemented
  int g()
  {
    printf("Enter Ext0::g\n");
    if (int err = Base::g())
      return err;
    printf("Enter Ext0::g\n");
    return 0;
  }
};

template <class Base>
class Ext1 : public Base
{
public:
  Ext1()
  {
    puts("Ext1::Ext1");
  }
  virtual ~Ext1()
  {
    puts("Ext1::~Ext1");
  }
  int f(int v, int &r, int &&rr, const int &cr)
  {
    printf("Enter Ext1::f(%d, %d, %d, %d)\n", v, r, rr, cr);
    if (int err = Base::f(v, r, std::move(rr), cr))
      return err;
    printf("Leave Ext1::f(%d, %d, %d, %d)\n", v, r, rr, cr);
    return 0;
  }
  // int g() is not implemented
};

template <class Base>
class VirtualExt : public Base
{
public:
  template<class ... Args>
  VirtualExt(Args &&...args)
    : Base(std::forward<Args>(args)...)
  {
    puts("VirtualExt::VirtualExt");
  }
  virtual ~VirtualExt()
  {
    puts("VirtualExt::~VirtualExt");
  }
  virtual int f(int v, int &r, int &&rr, const int &cr) = 0;
  virtual int g() = 0;
};

template <template <class Base> class ... Bases>
using CustomMix = Mix<VirtualExt, Bases..., Base0>;

class Base : public CustomMix<>
{};

class Derived : public CustomMix<Ext0, Ext1>
{
public:
  Derived()
  {
    puts("Derived::Derived");
  }
  virtual ~Derived()
  {
    puts("Derived::~Derived");
  }
  virtual int f(int v, int &r, int &&rr, const int &cr) override
  {
    printf("Enter Derived::f(%d, %d, %d, %d)\n", v, r, rr, cr);
    r = std::move(rr);
    rr = 5;
    printf("Leave Derived::f(%d, %d, %d, %d)\n", v, r, rr, cr);
    return 0;
  }
  virtual int g() override
  {
    printf("Derived::g()\n");
    return 0;
  }
  
};

int main()
{
  Derived d;
  int v = 1, r = 2, rr = 3, cr = 4;
  d.F(v, r, std::move(rr), cr);
  return 0;
}
